




#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <math.h>
#include <limits.h>

#if defined(ENABLE_DISPLAY)
 #include <GL/freeglut.h>
#endif

#include "body.h"
#include "bounding_volume.h"
#include "camera.h"
#include "file_reader.h"
#include "graphics.h"
#include "joint.h"
#include "robot.h"
#include "texture.h"

#include "vector3D.h"
#include "workspace.h"
#include "coldman.h"




static std::string PARAM_FILE_PATH;
static std::string PARAM_FILE;
static std::string TEXT;
int TEXT_X, TEXT_Y;


World world;




#if defined(ENABLE_DISPLAY)

static int WINDOW= 0;
static bool LEAVEMAINLOOP= false;
static bool WIREFRAME= false;
static bool SMOOTH= true;
static bool TEXTURES= false;
static bool SHADOWS= true;
static bool FILL= true;
static bool DRAW_JOINTS= false;
static unsigned int HELP= 0;
static unsigned int SELECTED_DOF= 0;
static unsigned int NB_DOFS= 0;

bool leftButton;
int mouseBeginX, mouseBeginY;

Camera camera;

TextureList texlist;

void setLight(GLfloat light_position[4]);
void setDimLight(GLfloat light_position[4]);
void setMaterial();
void setShadeMaterial();


//! Initialize OpenGL states and parameters.
//! \param width width of the display window
//! \param height height of the display window
void initViewerWindow(int width, int height)
{
  GLfloat light_position[4]= {world.lightPosition.x, world.lightPosition.y, world.lightPosition.z, 1.0};

  glColorMaterial(GL_FRONT,GL_AMBIENT_AND_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHTING);

  setLight(light_position);
  setMaterial();

  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);

  glDepthFunc(GL_LEQUAL);
  glEnable(GL_DEPTH_TEST);

  glShadeModel(GL_FLAT);
  glClearColor(0.5,0.6,1.0,0.0);

  glEnable(GL_LINE_SMOOTH);

  glEnable(GL_CULL_FACE);
  glCullFace(GL_BACK);
 
  setPerspectiveProjection(camera.fov, width, height);

  glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
  glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

  glEnable(GL_POLYGON_SMOOTH);
  glEnable(GL_LINE_SMOOTH);
}


//! GLUT keyboard callback for ordinary keys.
void keyboardCB(unsigned char key, int x, int y) 
{
  static bool displayGhosts= false;
  float delta= 0.15, factor= 0.1;

  switch(key) 
    {
    case 'z':
      camera.moveForward(delta);
      break;
    case 'Z':
      camera.moveForward(factor*delta);
      break;

    case 's':
      camera.moveForward(-delta);
      break;
    case 'S':
      camera.moveForward(-factor*delta);
      break;

    case 'q':
      camera.moveSideways(-delta);
      break;
    case 'Q':
      camera.moveSideways(-factor*delta);
      break;

    case 'd':
      camera.moveSideways(delta);
      break;
    case 'D':
      camera.moveSideways(factor*delta);
      break;

    case 'r':
      camera.move(0, 0, delta);
      break;
    case 'R':
      camera.move(0, 0, factor*delta);
      break;

    case 'f':
      camera.move(0, 0, -delta);
      break;
    case 'F':
      camera.move(0, 0, -factor*delta);
      break;
    case ' ':
      camera.switchMode();
      break;
    case 'X': case 27:
      LEAVEMAINLOOP= true;
      break;
    case 'w': case 'W':
      WIREFRAME= !WIREFRAME;
      if(WIREFRAME)
	{
	  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	}
      else
	{
	  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}
      break;
    case 'l': case 'L':
      SMOOTH= !SMOOTH;
      break;
    case 't': case 'T':
      TEXTURES= !TEXTURES;
      world.displayObstacleTextures(TEXTURES);
      break;
    case 'k': case 'K':
      SHADOWS= !SHADOWS;
      break;
    case 'v': case 'V':
      FILL= !FILL;
      break;
    case 'j': case 'J':
      DRAW_JOINTS= !DRAW_JOINTS;
      break;
    case 'h': case 'H':
      HELP= (HELP==0) ? 1 : 0;
      break;
    case '+':
      world.robot.incrementDOFvalue(SELECTED_DOF, 0.05);
      break;
    case '-':
      world.robot.decrementDOFvalue(SELECTED_DOF, 0.05);
      break;
    case 'p':
      world.robot.printConfiguration(DEGREE);
      break;
    case 'P':
      world.robot.printConfiguration(RADIAN);
      break;
    case 'g': case 'G':
      displayGhosts= !displayGhosts;
      if(displayGhosts)
	{  world.robot.enableDisplayOfCollisionBodiesOnly();  }
      else
	{  world.robot.resetDisplayStates();  }
      break;
    case 'i': case 'I':
      exportGLdisplay("screenshot.ppm");
      break;
    default: 
      break;
    }

  world.collisionTest();

  if(WINDOW!=0)
    {  glutPostRedisplay();  }
}

//! GLUT keyboard callback for special keys.
void specialCB(int key, int x, int y)
{
  float delta= 0.1;

  switch(key)
    {
    case GLUT_KEY_PAGE_UP:
      SELECTED_DOF++;
      if(SELECTED_DOF>NB_DOFS-1)
	SELECTED_DOF= NB_DOFS-1;
      break;
    case GLUT_KEY_PAGE_DOWN:
      SELECTED_DOF--;
      if(SELECTED_DOF==UINT_MAX)
	SELECTED_DOF= 0;
      break;
    case GLUT_KEY_UP:
      world.lightPosition.y+= delta; 
      break;
    case GLUT_KEY_DOWN:
      world.lightPosition.y-= delta; 
      break;
    case GLUT_KEY_LEFT:
      world.lightPosition.x+= delta; 
      break;
    case GLUT_KEY_RIGHT:
      world.lightPosition.x-= delta; 
      break;
    default: 
      break;
    }
  world.robot.selectDOF(SELECTED_DOF);

  if(WINDOW!=0)
    {  glutPostRedisplay();  }
}

//! GLUT mouse button callback.
void mouseCB(int _b, int _s, int _x, int _y)
{
  if (_s == GLUT_UP)
    {
      mouseBeginX= _x;
      mouseBeginY= _y; 
      return;
    }

  if (_b == GLUT_RIGHT_BUTTON)
    {
      leftButton= false;
      mouseBeginX= _x;
      mouseBeginY= _y; 
      return;
    }
  else
    {
      leftButton= true;
      mouseBeginX= _x;
      mouseBeginY= _y; 
    }
}

//! GLUT mouse motion callback.
void motionCB(int _x, int _y)
{
  float dx, dy;

  dx= (float) (_x - mouseBeginX);
  dy= (float) (_y - mouseBeginY);
  if(camera.mode==FREE_CAMERA)
    {
      camera.pan(dx/2.0);
      camera.tilt(-dy/2.0);
      mouseBeginX= _x;
      mouseBeginY= _y; 
    }
  else
    {
      if(leftButton)
	{
	  if( fabs(dx) > fabs(dy) )
	    {   
	      dx/= 3.0;
	      camera.changeAzimuth(-dx); 
	    }
	  else
	    { 
	      dy/= 3.0;
	      camera.changeElevation(dy);  
	    }

	}
      else
	{
	  camera.zoom(  -0.1*camera.distanceFromFocusPoint()*(dy/4.0)  );
	}
      mouseBeginX= _x;
      mouseBeginY= _y; 
    }
  
  if(WINDOW!=0)
    {  glutPostRedisplay();  }
}

//! GLUT mouse wheel callback.
void mouseWheelCB(int wheel, int direction, int x, int y)
{
  switch(direction)
    {
    case -1:
      camera.zoom(  0.15*camera.distanceFromFocusPoint() );
 
      break;
    case 1:
      camera.zoom(  -0.15*camera.distanceFromFocusPoint() );
      break;
    }

  if(WINDOW!=0)
    {  glutPostRedisplay();  }
}

//! Displays a character string on top of the 3D display, starting from a specified position ((0,0) is top-left corner)
//! \param str the string to display
//! \param x x-position of the text (in pixels)
//! \param y y-position of the text (in pixels)
void printText(const char *str, int x, int y)
{
  GLint viewport[4];
  glPushAttrib(GL_LIGHTING_BIT);

  glDisable(GL_LIGHTING);
  glColor3f(1.0, 1.0, 1.0);

  glGetIntegerv(GL_VIEWPORT, viewport);

  glPushMatrix();
  glRasterPos2i(x, viewport[3]-y);

  glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) str);
  glPopMatrix();

  glPopAttrib();
}

//! Displays the help text on top of the 3D display.
//! \param choice indicate which message to display
void printHelp(unsigned int choice)
{
  GLboolean lighting_enable;
  GLint viewport[4];

  lighting_enable= glIsEnabled(GL_LIGHTING);
  glGetIntegerv(GL_VIEWPORT, viewport);

  glDisable(GL_LIGHTING);
  glColor3f(1.0, 1.0, 1.0);
  glPushMatrix();
  glRasterPos2i(10, viewport[3]-30);
  switch(choice)
    {
    case 1:
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "-[esc, x]: quit.\n");
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "-[z, s, q, d, r, f]: move the camera\n");
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "  (forward, backward, left, right, up, down).\n");
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "-[space]: change the camera control mode.\n");
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "-[mouse left button]: rotate the camera\n");
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "  (depends on the selected mode).\n");
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "-[mouse right button/wheel]: zoom\n");
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "-[w]:toggle wireframe display mode.\n");
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "-[k]: toggle shadow display.\n");
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "-[t]: toggle texture display.\n");
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "-[l]: toggle smooth shading mode.\n");
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "-[j]: toggle robot's joint display.\n");
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "-[F1, F2, ...]: select a degree of freedom\n");
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "  of the robot.\n");
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "-[PAGE_UP, PAGE_DOWN]: select the next/previous\n");
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) " degree of freedom of the robot.\n");
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "-[+, -]: change the value\n");
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "  of the selected degree of freedom.\n");
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "-[ARROWS]: change the light position\n");
      break;
    default:
      glutBitmapString(GLUT_BITMAP_HELVETICA_12, (char unsigned *) "Press H for help.\n");
      break;
    }
  glPopMatrix();

  if(lighting_enable)
    {
      glEnable(GL_LIGHTING);
    }

}


//! Sets the default light parameters.
void setLight(GLfloat light_position[4])
{
  GLfloat light_ambient[4] = { 0.4f, 0.4f, 0.4f, 1.0f };
  GLfloat light_diffuse[4] = { 0.4f, 0.4f, 0.4f, 1.0f };
  GLfloat light_specular[4]= { 0.9f, 0.9f, 0.9f, 1.0f };

  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
  glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);

}

//! Sets the light parameters for the part of the floor that is in shadow.
void setDimLight(GLfloat light_position[4])
{
  GLfloat light_ambient[4] = { 0.1f, 0.1f, 0.1f, 1.0f };
  GLfloat light_diffuse[4] = { 0.1f, 0.1f, 0.1f, 1.0f };
  GLfloat light_specular[4]= { 0.0f, 0.0f, 0.0f, 1.0f };

  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
  glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);
}

//! Set the default material parameters.
void setMaterial()
{
  GLfloat mat_ambient[4]    = { 0.7f, 0.7f, 0.7f, 1.0f };
  GLfloat mat_diffuse[4]    = { 0.4f, 0.4f, 0.4f, 1.0f };
  GLfloat mat_specular[4]   = { 0.8f, 0.8f, 0.8f, 1.0f };
  GLfloat mat_emission[4]   = { 0.2f, 0.2f, 0.2f, 1.0f };
  GLfloat high_shininess = 100.0f;

  glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
  glMaterialfv(GL_FRONT, GL_EMISSION, mat_emission);
  glMaterialf(GL_FRONT, GL_SHININESS, high_shininess);
}

//! Set the material parameters for the shaded part of the floor.
void setShadeMaterial()
{
  GLfloat mat_ambient[4]    = { 0.2f, 0.2f, 0.2f, 1.0f };
  GLfloat mat_diffuse[4]    = { 0.2f, 0.2f, 0.2f, 1.0f };
  GLfloat mat_specular[4]   = { 0.2f, 0.2f, 0.2f, 1.0f };
  GLfloat mat_emission[4]   = { 0.2f, 0.2f, 0.2f, 1.0f };
  GLfloat high_shininess = 0.0f;


  glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
  glMaterialfv(GL_FRONT, GL_EMISSION, mat_emission);
  glMaterialf(GL_FRONT, GL_SHININESS, high_shininess);
}


//! GLUT display callback.
void displayCB()
{
  unsigned int i;
  GLdouble winX, winY, winZ;
  GLdouble modelMatrix[16], projMatrix[16];
  GLint viewport[4];
  float projection_matrix[16];
  float plane[4]= {0.0, 0.0, 1.0, -world.zmin};
  GLfloat light_position[4]= {world.lightPosition.x, world.lightPosition.y, world.lightPosition.z, 1.0};
  Vector3D pos;

  glGetIntegerv(GL_VIEWPORT, viewport);

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  setPerspectiveProjection(camera.fov, viewport[2], viewport[3]);

  camera.render();

  setLight(light_position);
  setMaterial();

  glDisable(GL_TEXTURE_2D);
  glEnable(GL_LIGHTING);
  glDisable(GL_CULL_FACE);

  //draw the world frame:
  HTMatrix I;
  I.setIdentity();
  drawFrame(I, 1.0);

  //draw the camera focus point:
  if(camera.mode==LOCKED_CAMERA)
    {
      glColor3f(0.0, 0.0, 0.0);
      glPushMatrix();
      glTranslatef(camera.focusPoint[0], camera.focusPoint[1], camera.focusPoint[2]);
      glutSolidSphere(0.01, 10, 10);
      glPopMatrix();
    }

  if(TEXTURES)
    {   glEnable(GL_TEXTURE_2D);   }
  else
    {   glDisable(GL_TEXTURE_2D);   }

  glShadeModel(GL_SMOOTH);

  setLight(light_position);

  if(SHADOWS)
    {   
      buildShadowMatrix(light_position, plane, projection_matrix);
  
      glClear(GL_STENCIL_BUFFER_BIT);
  
      glDepthFunc(GL_LEQUAL);
      glDisable(GL_STENCIL_TEST);
      glEnable(GL_LIGHT0);
  
      world.draw(SMOOTH);
  
      glEnable(GL_STENCIL_TEST);
      glStencilFunc(GL_ALWAYS, 0x2, 0x0);
      glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
  
      drawTiledFloor(1.0, 1.0, world.xmin, world.xmax, world.ymin, world.ymax, world.zmin, world.zmax);
  
      glDisable(GL_DEPTH_TEST);
      glMatrixMode(GL_MODELVIEW);
      glStencilFunc(GL_EQUAL, 0x3, 0x2);
      glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
      glPushMatrix();
      glMultMatrixf(projection_matrix);
      world.drawInBlack();
      glPopMatrix();
  
      glStencilFunc(GL_EQUAL, 0x1, 0x1);
      glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
      setDimLight(light_position);
      setShadeMaterial();
      drawTiledFloor(1.0, 1.0, world.xmin, world.xmax, world.ymin, world.ymax, world.zmin, world.zmax);
      glEnable(GL_DEPTH_TEST); 
      glDisable(GL_STENCIL_TEST);
    }
  else
    {  
      world.draw(!SMOOTH);
      drawTiledFloor(1.0, 1.0, world.xmin, world.xmax, world.ymin, world.ymax, world.zmin, world.zmax);
    }

  //glDisable(GL_LIGHTING);
  glDisable(GL_TEXTURE_2D);

  if(DRAW_JOINTS)
  { world.robot.drawJoints(0.1); }

  //Display the name of the joint corresponding to the selected DOF
  //(NB: it uses the OpenGL projection matrix to find where to print the name on the screen and
  //modifies the projection matrix to print the joint name):
  glGetDoublev(GL_MODELVIEW_MATRIX, modelMatrix);
  glGetDoublev(GL_PROJECTION_MATRIX, projMatrix);

  setOrthographicProjection(viewport[2], viewport[3]);
  glDisable(GL_LIGHTING);
  for(i=0; i<world.robot.joints.size(); ++i)
    {
      if(!world.robot.joints[i]->selected)
	{  continue;  }

      pos= world.robot.joints[i]->frame.fourthColumn();
      gluProject(pos.x, pos.y, pos.z, modelMatrix, projMatrix, viewport, &winX, &winY, &winZ);
      glColor3f(1.0, 0.0, 0.0);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      glRasterPos2d(winX, winY);
      glutBitmapString(GLUT_BITMAP_TIMES_ROMAN_24, (char unsigned *) world.robot.joints[i]->name().c_str());
      glPopMatrix();
      break;
    }

  if(HELP==0) {
    printText(TEXT.c_str(), TEXT_X, TEXT_Y);
  }
  printHelp(HELP); 


  glFlush();
  glutSwapBuffers();
}

//! GLUT reshape (window size modification) callback.
void reshapeCB(int width, int height)
{
  glViewport ( 0, 0, width, height ) ;
  setPerspectiveProjection(camera.fov, width, height);
  glMatrixMode(GL_MODELVIEW) ;
  glLoadIdentity() ;
}

//! function called by GLUT when the user clicks on the red "'x" button in the window header
void closeCB()
{
  WINDOW= 0;
}

void initGlutWindow()
{
  //  static int count= 0;
  //char str[128];

  //glutInitWindowPosition(600, 80); 
  //glutInitWindowSize(400, 300) ;

  glutInitWindowPosition(10, 10); 
  glutInitWindowSize(800, 600) ;
  glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH | GLUT_STENCIL);

  //create the window
  WINDOW= glutCreateWindow("COLlision Detection MANager");
  glutSetWindow(WINDOW);

  //set OpenGL graphics state -- material props, perspective, etc.
  initViewerWindow(800, 600);

  //set the callbacks
  glutDisplayFunc(displayCB);
  glutMouseFunc(mouseCB);
  glutMotionFunc(motionCB);  
  glutMouseWheelFunc(mouseWheelCB);
  glutKeyboardFunc(keyboardCB);
  glutSpecialFunc(specialCB) ;
  glutReshapeFunc(reshapeCB);
  glutCloseFunc(closeCB);

  //execution should continue after the user clicks on the "x" in the GLUT window header:
  glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_CONTINUE_EXECUTION);

  //GLUT should return control to the main program after the user clicks on the "x" in the window header:
  //glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

  //glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_EXIT);
 
}

#endif


extern "C" {


#if defined(ENABLE_DISPLAY)
  void coldmanDisplayInit() {
    int argc= 0;
    glutInit(&argc, NULL);
    initGlutWindow(); 
    return; 
  }
#endif
  int coldmanCheckCollision() {

    return world.earlyCollisionTest(0.0);
  }

  int coldmanPredictCollision(double time) {

    return world.earlyCollisionTest(time);
  }
#if defined(ENABLE_DISPLAY)
  void coldmanDisplayMainLoopEvent() {
    glutMainLoopEvent();
    return;
  }

  void coldmanDisplayRefresh() {
    glutPostRedisplay();
  }
#endif
  int coldmanEnvInit( char* path, char* file) {
    std::string _path;
    std::string _file;
    _path = path;
    _file = file;

    srand(time(NULL));

    if(world.loadFile(_path, _file)!=1){
      return 1;
    }
    return 0;
  }
#if defined(ENABLE_DISPLAY)
  int coldmanDisplayOn() {

    if(WINDOW!=0)
      {
	glutSetWindow(WINDOW);
	glutShowWindow();
	glutPostRedisplay();
      }
    else
      {
	initGlutWindow();
      }
    return 0;
  }

  int coldmanDisplayOff() {
    if(WINDOW==0)
    { return 1; }

    glutSetWindow(WINDOW);
    glutHideWindow();
    return 0;
  }

  int coldmanSetText(char *str, int x, int y) {
    if(str==NULL)
    { return 1; }

    TEXT= str; 
    TEXT_X= x;
    TEXT_Y= y;

    return 0;
  }
#endif
  int coldmanSetRobotQi(int dof, double value, double vel, double acc) {

    if(world.robot.setDOFvalue(dof, value)!=1) {
      return 1;
    }
    if(world.robot.setDOFvelocity(dof, vel)!=1) {
      return 1;
    }
    if(world.robot.setDOFacceleration(dof, acc)!=1) {
      return 1;
    }
    return 0;
  } 

  int coldmanDisableObstacleCollision(char *obstacle_name) {
    unsigned int i;
    std::string name= obstacle_name;    

    for(i=0; i<world.obstacles.size(); ++i) {
      if(world.obstacles[i]->name()==name) {
        world.obstacles[i]->enable_collision= false;
        world.obstacles[i]->colliding= false;
        coldmanCheckCollision(); //to update the colliding flags of bodies that may be in collision with the body to disable
        return 0;
      }
 
    }
    //obstacle is not defined:
    return 1;
  } 

  int coldmanEnableObstacleCollision(char *obstacle_name) {
    unsigned int i;
    std::string name= obstacle_name;    

    for(i=0; i<world.obstacles.size(); ++i) {
      if(world.obstacles[i]->name()==name) {
        if(world.obstacles[i]->type()==DISPLAY_ONLY || world.obstacles[i]->type()==NO_DISPLAY_NO_COLLISION) {
          //the obstacle type does not allow collision test (no collision model was created):
          return 1;
        }
        world.obstacles[i]->enable_collision= true;
        return 0;
      }
 
    }
    //the obstacle is not defined:
    return 1;
  }   

#if defined(ENABLE_DISPLAY)
  int coldmanDisableObstacleDisplay(char *obstacle_name) {
    unsigned int i;
    std::string name= obstacle_name;    

    for(i=0; i<world.obstacles.size(); ++i) {
      if(world.obstacles[i]->name()==name) {
        world.obstacles[i]->enable_display= false;
        coldmanDisplayRefresh();
        return 0;
      }
 
    }
    //obstacle is not defined:
    return 1;
  } 


  int coldmanEnableObstacleDisplay(char *obstacle_name) {
    unsigned int i;
    std::string name= obstacle_name;    

    for(i=0; i<world.obstacles.size(); ++i) {
      if(world.obstacles[i]->name()==name) {
        world.obstacles[i]->enable_display= true;
        coldmanDisplayRefresh();
        return 0;
      }
 
    }
    //the obstacle is not defined:
    return 1;
  }   

#endif

  int coldmanSetObstaclePosition(char *obstacle_name, double x, double y, double z) {
    unsigned int i;
    std::string name= obstacle_name;    

    for(i=0; i<world.obstacles.size(); ++i) {
      if(world.obstacles[i]->name()==name) {
         world.obstacles[i]->setPosition(x, y, z);
         return 0;
      }
    }
 
    //the obstacle is not defined:
    return 1;
  }  

  int coldmanSetObstacleOrientation(char *obstacle_name, double yaw, double pitch, double roll) {
    unsigned int i;
    std::string name= obstacle_name;    

    for(i=0; i<world.obstacles.size(); ++i) {
      if(world.obstacles[i]->name()==name) {
         world.obstacles[i]->setOrientation(yaw, pitch, roll);
         return 0;
      }
    }
 
    //the obstacle is not defined:
    return 1;
  }  

  int coldmanSetObstaclePose(char *obstacle_name,double x, double y, double z, double yaw, double pitch, double roll) {
    unsigned int i;
    std::string name= obstacle_name;    

    for(i=0; i<world.obstacles.size(); ++i) {
      if(world.obstacles[i]->name()==name) {
         world.obstacles[i]->setPose(x, y, z, yaw, pitch, roll);
         return 0;
      }
    }
 
    //the obstacle is not defined:
    return 1;
  }   

  int coldmanListAllObstacles() {
    unsigned int i;
    if(world.obstacles.empty()) {
      printf("There is no obstacle in the current world.\n");
      return 0;
    }

    for(i=0; i<world.obstacles.size(); ++i) {
      printf("obstacle #%d:  %s.\n", i+1, world.obstacles[i]->name().c_str());
    }
    return 0;
  }

#if defined(ENABLE_DISPLAY)

  //NB: angle is in degrees
  int coldmanChangeCameraAzimuth(double angle) {
    camera.changeAzimuth(angle);
    coldmanDisplayRefresh();
    return 0;
  }
  
  //NB: angle is in degrees
  int coldmanChangeCameraElevation(double angle) {
    camera.changeElevation(angle);
    coldmanDisplayRefresh();
    return 0;
  }

  int coldmanCameraZoom(double distance) {
    camera.zoom(-distance);
    coldmanDisplayRefresh();
    return 0;
  }

  int coldmanDisableShadows() {
    SHADOWS= false;
    coldmanDisplayRefresh();
    return 0;
  }

  int coldmanEnableShadows() {
    SHADOWS= true;
    coldmanDisplayRefresh();
    return 0;
  }

#endif

  int coldmanListCollidingBodies() {
    unsigned int i;

    for(i=0; i<world.robot.bodies.size(); ++i) {
      if(world.robot.bodies[i]->colliding==true) {
	//printf("  body \"%s\" of robot \"%s\" is colliding\n", i+1, world.robot.bodies[i]->name().c_str(), world.robot.name().c_str());
	printf("  body \"%s\" of robot \"%s\" is colliding\n", world.robot.bodies[i]->name().c_str(), world.robot.name().c_str());
      }
    }

    for(i=0; i<world.obstacles.size(); ++i) {
      if(world.obstacles[i]->colliding==true) {
	//  printf("  obstacle \"%s\" is colliding\n", i+1, world.obstacles[i]->name().c_str());
         printf("  obstacle \"%s\" is colliding\n",  world.obstacles[i]->name().c_str());
      }
    }
    return 0;
  }

#if defined(ENABLE_DISPLAY)

  int coldmanResizeWindow(int w, int h) {
    glutReshapeWindow(w, h);
    coldmanDisplayRefresh();
    return 0;
  }

  int coldmanPositionWindow(int x, int y) {
    glutPositionWindow(x, y);
    coldmanDisplayRefresh();
    return 0;
  }
#endif

  int coldmanLoadPointCloud(const char* filename)
  {
    unsigned int i;
    Body *body= NULL;
    HTMatrix pose;
    OccupancyGrid grid;
    std::vector<TriangleMesh> meshes;

    grid.loadFile(filename, 0.02);
    grid.labeling(10);
    grid.exportAsTriangleMeshes(meshes);

    pose.setIdentity();
 
    for(i=0; i<meshes.size(); ++i)
    {
       body= new Body("ScannedObstacle", DISPLAY_AND_COLLISION, meshes[i]);
       body->createPQPmodel();
       body->setColor(randomNumber(0.0,1.0), randomNumber(0.0,1.0), randomNumber(0.0,1.0), 1.0);
       world.addObstacle(body, pose);
    }

    return 0;
  }

  int coldmanRemoveScannedObstacles()
  {
     world.removeObstacles("ScannedObstacle");
     return 0;
  }


} 
/* END extern "C" */

